/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.1 Ganglinienprognose
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.glprog.glprog;

import java.lang.Thread.UncaughtExceptionHandler;
import java.text.MessageFormat;

import javax.persistence.EntityManagerFactory;

import com.bitctrl.util.jar.JarTools;
import com.bitctrl.util.logging.LoggerTools;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.iav.gllib.gllib.modell.GanglinienobjektFactory;
import de.bsvrz.iav.gllib.gllib.modell.db.GlSpeicher;
import de.bsvrz.iav.gllib.gllib.speicher.GlSpeicherUtil;
import de.bsvrz.sys.funclib.application.StandardApplication;
import de.bsvrz.sys.funclib.application.StandardApplicationRunner;
import de.bsvrz.sys.funclib.bitctrl.daf.LogTools;
import de.bsvrz.sys.funclib.bitctrl.modell.ObjektFactory;
import de.bsvrz.sys.funclib.bitctrl.modell.kalender.KalenderobjektFactory;
import de.bsvrz.sys.funclib.bitctrl.modell.systemmodellaoe.SystemModellAoeObjektFactory;
import de.bsvrz.sys.funclib.bitctrl.modell.systemmodellglobal.SystemModellGlobalObjektFactory;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.VerkehrsobjektFactory;
import de.bsvrz.sys.funclib.commandLineArgs.ArgumentList;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.operatingMessage.MessageSender;

/**
 * Die Datenverteilerapplikation <i>Ganglinienprognose</i>. Segment 5 Intelligente Analyseverfahren,
 * SWE 5.1 Ganglinienprognose
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 */
public final class GlProg implements StandardApplication, UncaughtExceptionHandler {

	/**
	 * Einsprungmethode der Applikation <i>Ganglinienprognose</i>.
	 *
	 * @param args
	 *            die Kommandozeilenargumente.
	 * @see #parseArguments(ArgumentList)
	 */
	public static void main(final String[] args) {
		StandardApplicationRunner.run(new GlProg(), args);
	}

	/** Logger dieser Klasse. */
	private Debug log;

	@SuppressWarnings("unused")
	private Verwaltung verwaltung;

	/** Ganglinien-Datenbank-IP. Standard: localhost. */
	private String dbIp;

	/** Ganglinien-Datenbank-Port. Standard: 1527. */
	private String dbPort;

	/** Ganglinien-Datenbank-Nutzer. Standard: derby. */
	private String dbUser;

	/** Ganglinien-Datenbank Passwort fuer Nutzer. Standard: derby. */
	private String dbPass;

	/** Ganglinien-Datenbank-Verzeichnis. Standard: gldb. */
	private String dbVerzeichnis;

	private int maxAnfragen;

	/**
	 * Initialisiert die Applikation.
	 */
	private GlProg() {
		JarTools.printVersionInfo(GlProg.class);
	}

	@Override
	public void initialize(final ClientDavInterface connection) throws Exception {
		MessageSender.getInstance().setApplicationLabel(
				getClass().getSimpleName() + "_" + connection.getLocalApplicationObject().getId());

		/*
		 * Workaround, weil das Archiv die erste Anfrage der Applikation manchmal nicht beantworten
		 * will.
		 */
		connection.getArchive();

		final ObjektFactory factory = ObjektFactory.getInstanz();
		factory.setVerbindung(connection);
		factory.registerFactory(new VerkehrsobjektFactory(), new KalenderobjektFactory(),
				new GanglinienobjektFactory(), new SystemModellAoeObjektFactory(),
				new SystemModellGlobalObjektFactory());

		final EntityManagerFactory emf = GlSpeicherUtil.getGlSpeicherEmf(dbIp, dbPort, dbUser,
				dbPass, dbVerzeichnis);
		GlSpeicher.init(emf);

		verwaltung = new Verwaltung(factory.getAOE(), maxAnfragen);

		LogTools.log(log, GlProgMsg.InfoGlProgBereit);
	}

	/**
	 * Als Parameter wird unter anderen die Datenbank angegeben. Folgende Parameter kennt die
	 * Ganglinienprognose:
	 *
	 * <ul>
	 * <li><code>-dbIp=localhost</code>: Der Hostname oder die IP-Adresse der Derby-Datenbank, in
	 * der die Ganglinien gespeichert werden.</li>
	 * <li><code>-dbPort=1527</code>: Der Port der Derby-Datenbank, in der die Ganglinien
	 * gespeichert werden.</li>
	 * <li><code>-dbUser=derby</code>: Der Nutzername der Derby-Datenbank, in der die Ganglinien
	 * gespeichert werden.</li>
	 * <li><code>-dbPass=derby</code>: Das Nutzerpasswort der Derby-Datenbank, in der die Ganglinien
	 * gespeichert werden.</li>
	 * <li><code>-dbVerzeichnis=gldb</code>: Der Name der Derby-Datenbank, in der die Ganglinien
	 * gespeichert werden.</li>
	 * <li><code>-maxAnfragen=20</code>: Die maximale Anzahl von Anfragen die gleichzeitig
	 * beantwortet werden sollen. Kommen mehr Anfragen, werden diese in eine Warteschlange gelegt
	 * nacheinander abgearbeitet.</li>
	 * </ul>
	 */
	@Override
	public void parseArguments(final ArgumentList argumentList) throws Exception {
		Thread.setDefaultUncaughtExceptionHandler(this);

		dbIp = argumentList.fetchArgument("-dbIp=localhost").asString(); //$NON-NLS-1$
		dbPort = argumentList.fetchArgument("-dbPort=1527").asString(); //$NON-NLS-1$
		dbUser = argumentList.fetchArgument("-dbUser=derby").asString(); //$NON-NLS-1$
		dbPass = argumentList.fetchArgument("-dbPass=derby").asString(); //$NON-NLS-1$
		dbVerzeichnis = argumentList.fetchArgument("-dbVerzeichnis=gldb") //$NON-NLS-1$
				.asString();
		maxAnfragen = Integer.parseInt(argumentList.fetchArgument("-maxAnfragen=20").asString()); //$NON-NLS-1$

		log = Debug.getLogger();
	}

	/**
	 * Gibt die Exception im Log und als Betriebsmeldung aus und beendet anschließend die
	 * Applikation.
	 */
	@Override
	public void uncaughtException(final Thread t, final Throwable e) {
		LogTools.log(log, GlProgMsg.ErrorUnbekannterFehler, t, LoggerTools.getStackTrace(e));
		if (log == null) {
			System.err.println(MessageFormat.format(GlProgMsg.ErrorUnbekannterFehler.toString(), t,
					LoggerTools.getStackTrace(e)));
		}
		System.exit(-1);
	}

}
