/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.1 Ganglinienprognose
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.glprog.glprog.util;

import static de.bsvrz.iav.gllib.gllib.dav.GanglinienMQOperationen.auschneiden;

import java.io.IOException;

import com.bitctrl.util.Interval;

import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.archive.ArchiveData;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKindCombination;
import de.bsvrz.dav.daf.main.archive.ArchiveDataQueryResult;
import de.bsvrz.dav.daf.main.archive.ArchiveDataSpecification;
import de.bsvrz.dav.daf.main.archive.ArchiveDataStream;
import de.bsvrz.dav.daf.main.archive.ArchiveOrder;
import de.bsvrz.dav.daf.main.archive.ArchiveQueryPriority;
import de.bsvrz.dav.daf.main.archive.ArchiveRequestManager;
import de.bsvrz.dav.daf.main.archive.ArchiveRequestOption;
import de.bsvrz.dav.daf.main.archive.ArchiveTimeSpecification;
import de.bsvrz.dav.daf.main.archive.TimingType;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.iav.gllib.gllib.dav.GanglinieMQ;
import de.bsvrz.iav.gllib.gllib.dav.GlProgAnfrage;
import de.bsvrz.iav.gllib.gllib.dav.Messwerte;
import de.bsvrz.iav.gllib.gllib.modell.parameter.PdGanglinienModellPrognose;
import de.bsvrz.iav.glprog.glprog.GlProgMsg;
import de.bsvrz.sys.funclib.bitctrl.daf.LogTools;
import de.bsvrz.sys.funclib.bitctrl.modell.ObjektFactory;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.MessQuerschnittAllgemein;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.onlinedaten.OdVerkehrsDatenKurzZeitMq;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.onlinedaten.OdVerkehrsDatenKurzZeitTrendExtraPolationMq;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Hilfsklasse für das Bestimmen der Analyseganglinie.
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 */
public final class AnalyseganglinieHelper {

	private GlProgAnfrage anfrage;
	private PdGanglinienModellPrognose.Daten prognoseParameter;

	/**
	 * Gibt die Anfrage zurück, zu der die Analyseganglinie bestimmt werden soll.
	 *
	 * @return die Anfrage.
	 */
	public GlProgAnfrage getAnfrage() {
		return anfrage;
	}

	/**
	 * Legt die Anfrage fest, zu der die Analyseganglinie bestimmt werden soll.
	 *
	 * @param anfrage
	 *            die Anfrage.
	 */
	public void setAnfrage(final GlProgAnfrage anfrage) {
		this.anfrage = anfrage;
	}

	/**
	 * Gibt die Parameter der Prognose zurück, die zur Bestimmung der
	 * Analyseganglinien verwendet werden.
	 *
	 * @return der Prognoseparameter.
	 */
	public PdGanglinienModellPrognose.Daten getPrognoseParameter() {
		return prognoseParameter;
	}

	/**
	 * Legt die Parameter der Prognose fest, die zur Bestimmung der
	 * Analyseganglinien verwendet werden.
	 *
	 * @param prognoseParameter
	 *            der Prognoseparameter.
	 */
	public void setPrognoseParameter(final PdGanglinienModellPrognose.Daten prognoseParameter) {
		this.prognoseParameter = prognoseParameter;
	}

	/**
	 * Erzeugt aus der Archivganglinie, den Kurzzeitdaten und den Daten der
	 * kurzfristigen Prognose die Analyseganglinie.
	 *
	 * @return die Analyseganglinie.
	 */
	public GanglinieMQ getAnalyseganglinie() {
		final OdVerkehrsDatenKurzZeitMq.Daten kurzzeitdaten = getKurzzeitdaten();
		final OdVerkehrsDatenKurzZeitTrendExtraPolationMq.Daten trendprognose = getTrendprognose();
		final GanglinieMQ archivGanglinie = getArchivdaten(anfrage, prognoseParameter, kurzzeitdaten);

		final GanglinieMQ analysegangline = archivGanglinie.clone();
		addKurzzeitdaten(analysegangline, kurzzeitdaten);
		addTrendprognose(analysegangline, trendprognose);

		return analysegangline;
	}

	/**
	 * Wenn es Kurzzeitdaten gibt, werden diese der Analyseganglinie hinzugefügt.
	 * 
	 * @param analysegangline
	 *            die Analyseganglinie
	 * @param kurzzeitdaten
	 *            die neuen Kurzzeitdaten
	 */
	private static void addKurzzeitdaten(final GanglinieMQ analysegangline,
			final OdVerkehrsDatenKurzZeitMq.Daten kurzzeitdaten) {
		if (!kurzzeitdaten.isValid()) {
			return;
		}

		final Messwerte messwert;
		final Double qKfz;
		final Double qLkw;
		final Double vPkw;
		final Double vLkw;

		if (kurzzeitdaten.getWert(OdVerkehrsDatenKurzZeitMq.Daten.Werte.QKfz.name()) != null) {
			qKfz = kurzzeitdaten.getWert(OdVerkehrsDatenKurzZeitMq.Daten.Werte.QKfz.name()).doubleValue();
		} else {
			qKfz = null;
		}
		if (kurzzeitdaten.getWert(OdVerkehrsDatenKurzZeitMq.Daten.Werte.QLkw.name()) != null) {
			qLkw = kurzzeitdaten.getWert(OdVerkehrsDatenKurzZeitMq.Daten.Werte.QLkw.name()).doubleValue();
		} else {
			qLkw = null;
		}
		if (kurzzeitdaten.getWert(OdVerkehrsDatenKurzZeitMq.Daten.Werte.VPkw.name()) != null) {
			vPkw = kurzzeitdaten.getWert(OdVerkehrsDatenKurzZeitMq.Daten.Werte.VPkw.name()).doubleValue();
		} else {
			vPkw = null;
		}
		if (kurzzeitdaten.getWert(OdVerkehrsDatenKurzZeitMq.Daten.Werte.VLkw.name()) != null) {
			vLkw = kurzzeitdaten.getWert(OdVerkehrsDatenKurzZeitMq.Daten.Werte.VLkw.name()).doubleValue();
		} else {
			vLkw = null;
		}
		messwert = new Messwerte(qKfz, qLkw, vPkw, vLkw);
		analysegangline.put(kurzzeitdaten.getZeitstempel(), messwert);
	}

	/**
	 * Wenn es eine Trendprognose der Kurzzeitdaten gibt, werden diese der
	 * Analyseganglinie hinzugefügt.
	 * 
	 * @param analysegangline
	 *            die Ganglinie
	 * @param trendprognose
	 *            die neuen Kurzzeitdaten
	 */
	private static void addTrendprognose(final GanglinieMQ analysegangline,
			final OdVerkehrsDatenKurzZeitTrendExtraPolationMq.Daten trendprognose) {
		if (!trendprognose.isValid()) {
			return;
		}

		final Messwerte messwert;
		final Double qKfz;
		final Double qLkw;
		final Double vPkw;
		final Double vLkw;

		if (trendprognose.getWert(OdVerkehrsDatenKurzZeitTrendExtraPolationMq.Daten.Werte.QKfzP.name()) != null) {
			qKfz = trendprognose.getWert(OdVerkehrsDatenKurzZeitTrendExtraPolationMq.Daten.Werte.QKfzP.name())
					.doubleValue();
		} else {
			qKfz = null;
		}
		if (trendprognose.getWert(OdVerkehrsDatenKurzZeitTrendExtraPolationMq.Daten.Werte.QLkwP.name()) != null) {
			qLkw = trendprognose.getWert(OdVerkehrsDatenKurzZeitTrendExtraPolationMq.Daten.Werte.QLkwP.name())
					.doubleValue();
		} else {
			qLkw = null;
		}
		if (trendprognose.getWert(OdVerkehrsDatenKurzZeitTrendExtraPolationMq.Daten.Werte.VPkwP.name()) != null) {
			vPkw = trendprognose.getWert(OdVerkehrsDatenKurzZeitTrendExtraPolationMq.Daten.Werte.VPkwP.name())
					.doubleValue();
		} else {
			vPkw = null;
		}
		if (trendprognose.getWert(OdVerkehrsDatenKurzZeitTrendExtraPolationMq.Daten.Werte.VLkwP.name()) != null) {
			vLkw = trendprognose.getWert(OdVerkehrsDatenKurzZeitTrendExtraPolationMq.Daten.Werte.VLkwP.name())
					.doubleValue();
		} else {
			vLkw = null;
		}
		messwert = new Messwerte(qKfz, qLkw, vPkw, vLkw);
		analysegangline.put(trendprognose.getZeitstempel(), messwert);
	}

	/**
	 * Fragt die aktuellen Kurzzeitdaten des Messquerschnitts vom Datenverteiler ab.
	 * Die Methode blockiert, bis die Kurzzeitdaten empfangen wurden.
	 *
	 * @return die aktuellen Kurzzeitdaten, aber niemals <code>null</code>.
	 */
	private OdVerkehrsDatenKurzZeitMq.Daten getKurzzeitdaten() {
		final MessQuerschnittAllgemein mq = anfrage.getMessQuerschnitt();
		final OdVerkehrsDatenKurzZeitMq odKurzzeitdaten = mq.getOnlineDatensatz(OdVerkehrsDatenKurzZeitMq.class);
		return odKurzzeitdaten.abrufenDatum(OdVerkehrsDatenKurzZeitMq.Aspekte.Analyse.getAspekt());
	}

	/**
	 * Fragt die aktuellen Trenddaten des Messquerschnitts vom Datenverteiler ab.
	 * Die Methode blockiert, bis die Trenddaten empfangen wurden.
	 *
	 * @return die aktuellen Trenddaten, aber niemals <code>null</code>.
	 */
	private OdVerkehrsDatenKurzZeitTrendExtraPolationMq.Daten getTrendprognose() {
		final MessQuerschnittAllgemein mq = anfrage.getMessQuerschnitt();
		final OdVerkehrsDatenKurzZeitTrendExtraPolationMq odTrend = mq
				.getOnlineDatensatz(OdVerkehrsDatenKurzZeitTrendExtraPolationMq.class);
		return odTrend.abrufenDatum(OdVerkehrsDatenKurzZeitTrendExtraPolationMq.Aspekte.PrognoseNormal.getAspekt());
	}

	/**
	 * Ruft die Archivdaten ab. Die Intervallgrenzen der Abfrage werden durch den
	 * Zeitstempel der Kurzzeitdaten und dem Prognoseparameter des Messquerschnitts
	 * bestimmt.
	 *
	 * <p>
	 * Die Synchronisierung dieser statischen Methode bewirkt, dass immer nur eine
	 * Archivanfrage gleichzeit stattfinden kann.
	 * 
	 * @param anfrage
	 *            die Anfrage
	 * @param prognoseParameter
	 *            die Parameter der Prognose
	 * @param kurzzeitdaten
	 *            die aktuellen Kurzzeitdaten.
	 * @return die Ganglinie, die aus den Archivdaten gebildet wurde.
	 */
	private synchronized static GanglinieMQ getArchivdaten(final GlProgAnfrage anfrage,
			final PdGanglinienModellPrognose.Daten prognoseParameter,
			final OdVerkehrsDatenKurzZeitMq.Daten kurzzeitdaten) {
		final MessQuerschnittAllgemein mq = anfrage.getMessQuerschnitt();
		final Debug log = Debug.getLogger();

		final GanglinieMQ archivGl = new GanglinieMQ();
		archivGl.setMessQuerschnitt(mq);

		final ArchiveRequestManager archiv = ObjektFactory.getInstanz().getVerbindung().getArchive();
		if (!archiv.isArchiveAvailable()) {
			LogTools.log(log, GlProgMsg.ErrorArchivNichtVerfuegbar);
			return archivGl;
		}

		// Anfrage zusammenstellen
		final long patternMatchingOffset = prognoseParameter.getPatternMatchingOffset();
		final Interval intervall;
		if (kurzzeitdaten.isValid()) {
			// Archivdaten abrufen vom Zeitpunkt der Kurzzeitdaten bis
			// Matching-Offset in die Vergangenheit.
			intervall = new Interval(kurzzeitdaten.getZeitstempel() - patternMatchingOffset,
					kurzzeitdaten.getZeitstempel());
		} else {
			// Archivdaten abrufen vom Startzeitpunkt der Anfrage bis
			// Matching-Offset in die Vergangenheit.
			final long startPrognoseZeitraum = anfrage.getPrognoseZeitraum().getStart();
			intervall = new Interval(startPrognoseZeitraum - patternMatchingOffset, startPrognoseZeitraum);
		}
		final DataModel modell = ObjektFactory.getInstanz().getVerbindung().getDataModel();
		final AttributeGroup atg = modell.getAttributeGroup("atg.verkehrsDatenKurzZeitMq"); //$NON-NLS-1$
		final Aspect asp = modell.getAspect("asp.analyse"); //$NON-NLS-1$
		final DataDescription dbs = new DataDescription(atg, asp);
		final ArchiveTimeSpecification timeSpec = new ArchiveTimeSpecification(TimingType.DATA_TIME, false,
				intervall.getStart(), intervall.getEnd());
		final ArchiveDataKindCombination dataKinds = new ArchiveDataKindCombination(ArchiveDataKind.ONLINE);
		final ArchiveDataSpecification spec = new ArchiveDataSpecification(timeSpec, dataKinds,
				ArchiveOrder.BY_DATA_TIME, ArchiveRequestOption.NORMAL, dbs, mq.getSystemObject());

		// Anfrage verschicken und auf Antwort warten
		final ArchiveDataQueryResult antwort = archiv.request(ArchiveQueryPriority.MEDIUM, spec);

		try {
			assert antwort.getStreams().length == 1;
			final ArchiveDataStream strom = antwort.getStreams()[0];

			// Daten auslesen und Ganglinie daraus erstellen
			while (true) {
				ArchiveData archivdaten;
				Data daten;
				long timestamp;
				Double qKfz = null;
				Double qLkw = null;
				Double vLkw = null;
				Double vPkw = null;

				archivdaten = strom.take();
				if (archivdaten == null) {
					// Datenstrom ist zu Ende
					break;
				}

				timestamp = archivdaten.getDataTime();
				daten = archivdaten.getData();
				if (daten == null) {
					continue; // ein leerer Datensatz wird ignoriert
				}
				if (!daten.getItem("QKfz").getScaledValue("Wert").isState()) { //$NON-NLS-1$//$NON-NLS-2$
					qKfz = daten.getItem("QKfz").getScaledValue("Wert") //$NON-NLS-1$ //$NON-NLS-2$
							.doubleValue();
				}
				if (!daten.getItem("QLkw").getScaledValue("Wert").isState()) { //$NON-NLS-1$//$NON-NLS-2$
					qLkw = daten.getItem("QLkw").getScaledValue("Wert") //$NON-NLS-1$ //$NON-NLS-2$
							.doubleValue();
				}
				if (!daten.getItem("VLkw").getScaledValue("Wert").isState()) { //$NON-NLS-1$//$NON-NLS-2$
					vLkw = daten.getItem("VLkw").getScaledValue("Wert") //$NON-NLS-1$ //$NON-NLS-2$
							.doubleValue();
				}
				if (!daten.getItem("VPkw").getScaledValue("Wert").isState()) { //$NON-NLS-1$//$NON-NLS-2$
					vPkw = daten.getItem("VPkw").getScaledValue("Wert") //$NON-NLS-1$ //$NON-NLS-2$
							.doubleValue();
				}
				archivGl.put(timestamp, new Messwerte(qKfz, qLkw, vPkw, vLkw));
			}
		} catch (final IllegalStateException ex) {
			archivGl.clear();
			LogTools.log(log, GlProgMsg.ErrorBeiArchivanfrage, ex.getLocalizedMessage());
		} catch (final InterruptedException ex) {
			archivGl.clear();
			LogTools.log(log, GlProgMsg.ErrorBeiArchivanfrage, ex.getLocalizedMessage());
		} catch (final IOException ex) {
			archivGl.clear();
			LogTools.log(log, GlProgMsg.ErrorBeiArchivanfrage, ex.getLocalizedMessage());
		}

		if (archivGl.getStuetzstellen().size() > 0) {
			auschneiden(archivGl, intervall);
			LogTools.log(log, GlProgMsg.FinerArchivganglinie, archivGl);
		}

		return archivGl;
	}

}
